@@ -14,17 +14,33 @@
 #include <lib/ilist.h>
 #include <mb/pg_wchar.h>
 #include <miscadmin.h>
+#include <optimizer/cost.h>
+#include <storage/bufmgr.h>
 #include <storage/ipc.h>
 #include <storage/lmgr.h>
 #include <utils/array.h>
 #include <utils/builtins.h>
 #include <utils/lsyscache.h>
 #include <utils/selfuncs.h>
+#include <utils/snapmgr.h>
+#include <utils/timestamp.h>
+#include <utils/tqual.h>
 #include <utils/typcache.h>
 
 #include <groonga.h>
 
 #include <stdlib.h>
+#include <math.h>
+#include <sys/types.h>
+#include <sys/stat.h>
+
+#ifdef WIN32
+typedef struct _stat pgrn_stat_buffer;
+#    define pgrn_stat(path, buffer) _stat(path, buffer)
+#else
+typedef struct stat pgrn_stat_buffer;
+#    define pgrn_stat(path, buffer) stat(path, buffer)
+#endif
 
 #define VARCHARARRAYOID 1015
 
@@ -44,6 +60,7 @@ typedef struct PGrnCreateData
 {
 	Relation index;
 	grn_obj *sourcesTable;
+	grn_obj *sourcesCtidColumn;
 	grn_obj *lexicons;
 	unsigned int i;
 	TupleDesc desc;
@@ -56,6 +73,7 @@ typedef struct PGrnCreateData
 typedef struct PGrnBuildStateData
 {
 	grn_obj	*sourcesTable;
+	grn_obj	*sourcesCtidColumn;
 	double nIndexedTuples;
 } PGrnBuildStateData;
 
@@ -65,7 +83,17 @@ typedef struct PGrnScanOpaqueData
 {
 	slist_node node;
 	Oid dataTableID;
+	struct
+	{
+		AttrNumber number;
+		Oid type;
+		grn_id domain;
+		unsigned char flags;
+		grn_obj *lexicon;
+		grn_obj *indexColumn;
+	} primaryKey;
 	grn_obj *sourcesTable;
+	grn_obj *sourcesCtidColumn;
 	grn_obj minBorderValue;
 	grn_obj maxBorderValue;
 	grn_obj *searched;
@@ -73,7 +101,7 @@ typedef struct PGrnScanOpaqueData
 	grn_obj *targetTable;
 	grn_obj *indexCursor;
 	grn_table_cursor *tableCursor;
-	grn_obj *keyAccessor;
+	grn_obj *ctidAccessor;
 	grn_obj *scoreAccessor;
 	grn_id currentID;
 } PGrnScanOpaqueData;
@@ -118,6 +146,8 @@ PG_FUNCTION_INFO_V1(pgroonga_options);
 static grn_ctx grnContext;
 static grn_ctx *ctx = &grnContext;
 static grn_obj buffer;
+static grn_obj ctidBuffer;
+static grn_obj scoreBuffer;
 static grn_obj headBuffer;
 static grn_obj bodyBuffer;
 static grn_obj footBuffer;
@@ -165,24 +195,67 @@ PGrnEnsureDatabase(void)
 {
 	char path[MAXPGPATH];
 	grn_obj	*db;
+	pgrn_stat_buffer file_status;
 
 	GRN_CTX_SET_ENCODING(ctx, PGrnGetEncoding());
 	join_path_components(path,
 						 GetDatabasePath(MyDatabaseId, DEFAULTTABLESPACE_OID),
 						 PGrnDatabaseBasename);
 
-	db = grn_db_open(ctx, path);
-	if (db)
-		return;
-
-	db = grn_db_create(ctx, path, NULL);
-	if (!db)
-		ereport(ERROR,
-				(errcode(ERRCODE_IO_ERROR),
-				 errmsg("pgroonga: failed to create database: <%s>: %s",
-						path, ctx->errbuf)));
+	if (pgrn_stat(path, &file_status) == 0)
+	{
+		db = grn_db_open(ctx, path);
+		if (!db)
+			ereport(ERROR,
+					(errcode(ERRCODE_IO_ERROR),
+					 errmsg("pgroonga: failed to open database: <%s>: %s",
+							path, ctx->errbuf)));
+	}
+	else
+	{
+		db = grn_db_create(ctx, path, NULL);
+		if (!db)
+			ereport(ERROR,
+					(errcode(ERRCODE_IO_ERROR),
+					 errmsg("pgroonga: failed to create database: <%s>: %s",
+							path, ctx->errbuf)));
+	}
 }
 
+static void
+PGrnLoggerLog(grn_ctx *ctx,
+			  grn_log_level level,
+			  const char *timestamp,
+			  const char *title,
+			  const char *message,
+			  const char *location,
+			  void *userData)
+{
+	const char levelTags[] = " EACewnid-";
+
+	if (location && *location)
+	{
+		ereport(LOG,
+				(errmsg("pgroonga: %s|%c|%s %s %s",
+						timestamp, levelTags[level], title, message, location)));
+	}
+	else
+	{
+		ereport(LOG,
+				(errmsg("pgroonga: %s|%c|%s %s",
+						timestamp, levelTags[level], title, message)));
+	}
+}
+
+static grn_logger PGrnLogger = {
+	GRN_LOG_DEFAULT_LEVEL,
+	GRN_LOG_TIME | GRN_LOG_MESSAGE,
+	NULL,
+	PGrnLoggerLog,
+	NULL,
+	NULL
+};
+
 static void
 PGrnOnProcExit(int code, Datum arg)
 {
@@ -192,6 +265,8 @@ PGrnOnProcExit(int code, Datum arg)
 	GRN_OBJ_FIN(ctx, &footBuffer);
 	GRN_OBJ_FIN(ctx, &bodyBuffer);
 	GRN_OBJ_FIN(ctx, &headBuffer);
+	GRN_OBJ_FIN(ctx, &ctidBuffer);
+	GRN_OBJ_FIN(ctx, &scoreBuffer);
 	GRN_OBJ_FIN(ctx, &buffer);
 
 	db = grn_ctx_db(ctx);
@@ -322,9 +397,13 @@ _PG_init(void)
 				(errcode(ERRCODE_SYSTEM_ERROR),
 				 errmsg("pgroonga: failed to initialize Groonga context")));
 
+	grn_logger_set(ctx, &PGrnLogger);
+
 	on_proc_exit(PGrnOnProcExit, 0);
 
 	GRN_VOID_INIT(&buffer);
+	GRN_FLOAT_INIT(&scoreBuffer, 0);
+	GRN_UINT64_INIT(&ctidBuffer, 0);
 	GRN_TEXT_INIT(&headBuffer, 0);
 	GRN_TEXT_INIT(&bodyBuffer, 0);
 	GRN_TEXT_INIT(&footBuffer, 0);
@@ -461,25 +540,123 @@ PGrnGetType(Relation index, AttrNumber n, unsigned char *flags)
 }
 
 static void
-PGrnGetValue(Relation index, AttrNumber n, grn_obj *buffer, Datum value)
+PGrnConvertDatumArrayType(Datum datum, Oid typeID, grn_obj *buffer)
 {
-	FmgrInfo *function;
+	ArrayType *value = DatumGetArrayTypeP(datum);
+	int i, n;
 
-	function = index_getprocinfo(index, n + 1, PGrnGetValueProc);
-	FunctionCall3(function,
-				  PointerGetDatum(ctx), PointerGetDatum(buffer),
-				  value);
+	n = ARR_DIMS(value)[0];
+	for (i = 1; i <= n; i++)
+	{
+		int weight = 0;
+		Datum elementDatum;
+		VarChar *element;
+		bool isNULL;
+
+		elementDatum = array_ref(value, 1, &i, -1, -1, false, 'i', &isNULL);
+		if (isNULL)
+			continue;
+
+		switch (typeID)
+		{
+		case VARCHARARRAYOID:
+			element = DatumGetVarCharPP(elementDatum);
+			grn_vector_add_element(ctx, buffer,
+								   VARDATA_ANY(element),
+								   VARSIZE_ANY_EXHDR(element),
+								   weight,
+								   buffer->header.domain);
+			break;
+		case TEXTARRAYOID:
+			element = DatumGetTextPP(elementDatum);
+			grn_vector_add_element(ctx, buffer,
+								   VARDATA_ANY(element),
+								   VARSIZE_ANY_EXHDR(element),
+								   weight,
+								   buffer->header.domain);
+			break;
+		}
+	}
 }
 
 static void
-PGrnGetQuery(Relation index, AttrNumber n, grn_obj *buffer, Datum value)
+PGrnConvertDatum(Datum datum, Oid typeID, grn_obj *buffer)
 {
-	FmgrInfo *function;
+	switch (typeID)
+	{
+	case BOOLOID:
+		GRN_BOOL_SET(ctx, buffer, DatumGetBool(datum));
+		break;
+	case INT2OID:
+		GRN_INT16_SET(ctx, buffer, DatumGetInt16(datum));
+		break;
+	case INT4OID:
+		GRN_INT32_SET(ctx, buffer, DatumGetInt32(datum));
+		break;
+	case INT8OID:
+		GRN_INT64_SET(ctx, buffer, DatumGetInt64(datum));
+		break;
+	case FLOAT4OID:
+		GRN_FLOAT_SET(ctx, buffer, DatumGetFloat4(datum));
+		break;
+	case FLOAT8OID:
+		GRN_FLOAT_SET(ctx, buffer, DatumGetFloat8(datum));
+		break;
+	case TIMESTAMPOID:
+	case TIMESTAMPTZOID:
+	{
+		Timestamp value = DatumGetTimestamp(datum);
+		pg_time_t unixTime;
+		int32 usec;
 
-	function = index_getprocinfo(index, n + 1, PGrnGetQueryProc);
-	FunctionCall3(function,
-				  PointerGetDatum(ctx), PointerGetDatum(buffer),
-				  value);
+		unixTime = timestamptz_to_time_t(value);
+#ifdef HAVE_INT64_TIMESTAMP
+		usec = value % USECS_PER_SEC;
+#else
+		{
+			double rawUsec;
+			modf(value, &rawUsec);
+			usec = rawUsec * USECS_PER_SEC;
+			if (usec < 0.0)
+			{
+				usec = -usec;
+			}
+		}
+#endif
+		GRN_TIME_SET(ctx, buffer, GRN_TIME_PACK(unixTime, usec));
+		break;
+	}
+	case TEXTOID:
+	case XMLOID:
+	{
+		text *value = DatumGetTextPP(datum);
+		GRN_TEXT_SET(ctx, buffer,
+					 VARDATA_ANY(value), VARSIZE_ANY_EXHDR(value));
+		break;
+	}
+	case VARCHAROID:
+	{
+		VarChar *value = DatumGetVarCharPP(datum);
+		GRN_TEXT_SET(ctx, buffer,
+					 VARDATA_ANY(value), VARSIZE_ANY_EXHDR(value));
+		break;
+	}
+#ifdef NOT_USED
+	case POINTOID:
+		/* GRN_DB_TOKYO_GEO_POINT or GRN_DB_WGS84_GEO_POINT; */
+		break;
+#endif
+	case VARCHARARRAYOID:
+	case TEXTARRAYOID:
+		PGrnConvertDatumArrayType(datum, typeID, buffer);
+		break;
+	default:
+		ereport(ERROR,
+				(errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
+				 errmsg("pgroonga: unsupported datum type: %u",
+						typeID)));
+		break;
+	}
 }
 
 static grn_obj *
@@ -498,8 +675,19 @@ PGrnLookupSourcesTable(Relation index, int errorLevel)
 {
 	char name[GRN_TABLE_MAX_KEY_SIZE];
 
-	snprintf(name, sizeof(name)
-			 , PGrnSourcesTableNameFormat,
+	snprintf(name, sizeof(name),
+			 PGrnSourcesTableNameFormat,
+			 index->rd_node.relNode);
+	return PGrnLookup(name, errorLevel);
+}
+
+static grn_obj *
+PGrnLookupSourcesCtidColumn(Relation index, int errorLevel)
+{
+	char name[GRN_TABLE_MAX_KEY_SIZE];
+
+	snprintf(name, sizeof(name),
+			 PGrnSourcesTableNameFormat "." PGrnSourcesCtidColumnName,
 			 index->rd_node.relNode);
 	return PGrnLookup(name, errorLevel);
 }
@@ -578,6 +766,29 @@ PGrnIsForFullTextSearchIndex(Relation index, int nthAttribute)
 	return OidIsValid(queryStrategyOID);
 }
 
+static void
+PGrnCreateSourcesCtidColumn(PGrnCreateData *data)
+{
+	data->sourcesCtidColumn = PGrnCreateColumn(data->sourcesTable,
+											   PGrnSourcesCtidColumnName,
+											   GRN_OBJ_COLUMN_SCALAR,
+											   grn_ctx_at(ctx, GRN_DB_UINT64));
+}
+
+static void
+PGrnCreateSourcesTable(PGrnCreateData *data)
+{
+	char sourcesTableName[GRN_TABLE_MAX_KEY_SIZE];
+
+	snprintf(sourcesTableName, sizeof(sourcesTableName),
+			 PGrnSourcesTableNameFormat, data->relNode);
+	data->sourcesTable = PGrnCreateTable(sourcesTableName,
+										 GRN_OBJ_TABLE_NO_KEY,
+										 NULL);
+
+	PGrnCreateSourcesCtidColumn(data);
+}
+
 static void
 PGrnCreateDataColumn(PGrnCreateData *data)
 {
@@ -677,23 +888,22 @@ PGrnCreateIndexColumn(PGrnCreateData *data)
  * @param	index
  */
 static void
-PGrnCreate(Relation index, grn_obj **sourcesTable, grn_obj *lexicons)
+PGrnCreate(Relation index,
+		   grn_obj **sourcesTable,
+		   grn_obj **sourcesCtidColumn,
+		   grn_obj *lexicons)
 {
-	char sourcesTableName[GRN_TABLE_MAX_KEY_SIZE];
 	PGrnCreateData data;
 
 	data.index = index;
 	data.desc = RelationGetDescr(index);
 	data.relNode = index->rd_node.relNode;
-
-	snprintf(sourcesTableName, sizeof(sourcesTableName),
-			 PGrnSourcesTableNameFormat, data.relNode);
-	*sourcesTable = PGrnCreateTable(sourcesTableName,
-									GRN_OBJ_TABLE_PAT_KEY,
-									grn_ctx_at(ctx, GRN_DB_UINT64));
-	data.sourcesTable = *sourcesTable;
 	data.lexicons = lexicons;
 
+	PGrnCreateSourcesTable(&data);
+	*sourcesTable = data.sourcesTable;
+	*sourcesCtidColumn = data.sourcesCtidColumn;
+
 	for (data.i = 0; data.i < data.desc->natts; data.i++)
 	{
 		data.forFullTextSearch = PGrnIsForFullTextSearchIndex(index, data.i);
@@ -734,7 +944,7 @@ PGrnSetSources(Relation index, grn_obj *sourcesTable)
 	GRN_OBJ_FIN(ctx, &sourceIDs);
 }
 
-static grn_id
+static uint64
 CtidToUInt64(ItemPointer ctid)
 {
 	BlockNumber blockNumber;
@@ -742,7 +952,7 @@ CtidToUInt64(ItemPointer ctid)
 
 	blockNumber = ItemPointerGetBlockNumber(ctid);
 	offsetNumber = ItemPointerGetOffsetNumber(ctid);
-	return (blockNumber << 16 | offsetNumber);
+	return (((uint64)blockNumber << 16) | ((uint64)offsetNumber));
 }
 
 static ItemPointerData
@@ -753,40 +963,97 @@ UInt64ToCtid(uint64 key)
 	return ctid;
 }
 
+static bool
+PGrnIsAliveCtid(Relation table, ItemPointer ctid)
+{
+	Buffer buffer;
+	HeapTupleData tuple;
+	Snapshot snapshot;
+	ItemPointerData realCtid;
+	bool allDead;
+	bool found;
+	bool isAlive = false;
+
+	buffer = ReadBuffer(table, ItemPointerGetBlockNumber(ctid));
+	snapshot = RegisterSnapshot(GetLatestSnapshot());
+	realCtid = *ctid;
+	found = heap_hot_search_buffer(&realCtid, table, buffer, snapshot, &tuple,
+								   &allDead, true);
+	if (found) {
+		uint64 tupleID;
+
+		tupleID = CtidToUInt64(&(tuple.t_self));
+		isAlive = (tupleID == CtidToUInt64(ctid) ||
+				   tupleID == CtidToUInt64(&realCtid));
+	}
+	UnregisterSnapshot(snapshot);
+	ReleaseBuffer(buffer);
+
+	return isAlive;
+}
+
 static double
-PGrnCollectScore(Oid tableID, ItemPointer ctid)
+PGrnCollectScoreScanOpaque(Relation table, HeapTuple tuple, PGrnScanOpaque so)
 {
 	double score = 0.0;
-	uint64_t key;
-	grn_id recordID = GRN_ID_NIL;
-	slist_iter iter;
-	grn_obj scoreBuffer;
+	TupleDesc desc;
+	bool isNULL;
+	Datum primaryKeyValue;
+	grn_table_cursor *tableCursor;
+	grn_obj *indexCursor;
+	grn_id recordID;
 
-	key = CtidToUInt64(ctid);
-	GRN_FLOAT_INIT(&scoreBuffer, 0);
+	if (so->dataTableID != tuple->t_tableOid)
+		return 0.0;
 
-	slist_foreach(iter, &PGrnScanOpaques)
+	if (!so->scoreAccessor)
+		return 0.0;
+
+	if (!OidIsValid(so->primaryKey.type))
+		return 0.0;
+
+	grn_obj_reinit(ctx, &buffer, so->primaryKey.domain, so->primaryKey.flags);
+
+	desc = RelationGetDescr(table);
+	primaryKeyValue = heap_getattr(tuple, so->primaryKey.number, desc, &isNULL);
+	PGrnConvertDatum(primaryKeyValue, so->primaryKey.type, &buffer);
+
+	tableCursor = grn_table_cursor_open(ctx, so->primaryKey.lexicon,
+										GRN_BULK_HEAD(&buffer),
+										GRN_BULK_VSIZE(&buffer),
+										GRN_BULK_HEAD(&buffer),
+										GRN_BULK_VSIZE(&buffer),
+										0, -1, GRN_CURSOR_ASCENDING);
+	if (!tableCursor)
+		return 0.0;
+
+
+	indexCursor = grn_index_cursor_open(ctx,
+										tableCursor,
+										so->primaryKey.indexColumn,
+										GRN_ID_NIL, GRN_ID_MAX, 0);
+	if (!indexCursor)
+	{
+		grn_table_cursor_close(ctx, tableCursor);
+		return 0.0;
+	}
+
+	while ((recordID = grn_table_cursor_next(ctx, indexCursor)) != GRN_ID_NIL)
 	{
-		PGrnScanOpaque so;
 		grn_id id;
-
-		so = slist_container(PGrnScanOpaqueData, node, iter.cur);
-		if (so->dataTableID != tableID)
-			continue;
-
-		if (!so->scoreAccessor)
-			continue;
-
-		if (recordID == GRN_ID_NIL)
-		{
-			recordID = grn_table_get(ctx, so->sourcesTable,
-									 &key, sizeof(uint64_t));
-		}
+		ItemPointerData ctid;
 
 		id = grn_table_get(ctx, so->searched, &recordID, sizeof(grn_id));
 		if (id == GRN_ID_NIL)
 			continue;
 
+		GRN_BULK_REWIND(&ctidBuffer);
+		grn_obj_get_value(ctx, so->ctidAccessor, id, &ctidBuffer);
+		ctid = UInt64ToCtid(GRN_UINT64_VALUE(&ctidBuffer));
+
+		if (!PGrnIsAliveCtid(table, &ctid))
+			continue;
+
 		GRN_BULK_REWIND(&scoreBuffer);
 		grn_obj_get_value(ctx, so->scoreAccessor, id, &scoreBuffer);
 		if (scoreBuffer.header.domain == GRN_DB_FLOAT)
@@ -799,7 +1066,25 @@ PGrnCollectScore(Oid tableID, ItemPointer ctid)
 		}
 	}
 
-	GRN_OBJ_FIN(ctx, &scoreBuffer);
+	grn_obj_unlink(ctx, indexCursor);
+	grn_obj_unlink(ctx, tableCursor);
+
+	return score;
+}
+
+static double
+PGrnCollectScore(Relation table, HeapTuple tuple)
+{
+	double score = 0.0;
+	slist_iter iter;
+
+	slist_foreach(iter, &PGrnScanOpaques)
+	{
+		PGrnScanOpaque so;
+
+		so = slist_container(PGrnScanOpaqueData, node, iter.cur);
+		score += PGrnCollectScoreScanOpaque(table, tuple, so);
+	}
 
 	return score;
 }
@@ -820,10 +1105,22 @@ pgroonga_score(PG_FUNCTION_ARGS)
 	recordType = HeapTupleHeaderGetTypMod(header);
 	desc = lookup_rowtype_tupdesc(type, recordType);
 
-	if (desc->natts > 0)
+	if (desc->natts > 0 && !slist_is_empty(&PGrnScanOpaques))
 	{
-		Oid tableID = desc->attrs[0]->attrelid;
-		score = PGrnCollectScore(tableID, &(header->t_ctid));
+		HeapTupleData tupleData;
+		HeapTuple tuple;
+		Relation table;
+
+		tupleData.t_len = HeapTupleHeaderGetDatumLength(header);
+		tupleData.t_tableOid = desc->attrs[0]->attrelid;
+		tupleData.t_data = header;
+		tuple = &tupleData;
+
+		table = RelationIdGetRelation(tuple->t_tableOid);
+
+		score = PGrnCollectScore(table, tuple);
+
+		RelationClose(table);
 	}
 
 	ReleaseTupleDesc(desc);
@@ -907,38 +1204,23 @@ pgroonga_command(PG_FUNCTION_ARGS)
 }
 
 static grn_bool
-pgroonga_contain_raw(const char *text, unsigned int text_size,
-					 const char *key, unsigned int key_size)
+pgroonga_contain_raw(const char *text, unsigned int textSize,
+					 const char *subText, unsigned int subTextSize)
 {
-	grn_bool contained = GRN_FALSE;
-	grn_obj buffer;
-	grn_obj *expression, *expressionVariable;
+	grn_bool contained;
+	grn_obj targetBuffer;
+	grn_obj subTextBuffer;
 
-	GRN_EXPR_CREATE_FOR_QUERY(ctx, NULL, expression, expressionVariable);
+	GRN_TEXT_INIT(&targetBuffer, GRN_OBJ_DO_SHALLOW_COPY);
+	GRN_TEXT_SET(ctx, &targetBuffer, text, textSize);
 
-	GRN_TEXT_INIT(&buffer, 0);
+	GRN_TEXT_INIT(&subTextBuffer, GRN_OBJ_DO_SHALLOW_COPY);
+	GRN_TEXT_SET(ctx, &subTextBuffer, subText, subTextSize);
 
-	GRN_TEXT_SET(ctx, &buffer, text, text_size);
-	grn_expr_append_const(ctx, expression, &buffer, GRN_OP_PUSH, 1);
+	contained = grn_operator_exec_match(ctx, &targetBuffer, &subTextBuffer);
 
-	GRN_TEXT_SET(ctx, &buffer, key, key_size);
-	grn_expr_append_const(ctx, expression, &buffer, GRN_OP_PUSH, 1);
-
-	grn_expr_append_op(ctx, expression, GRN_OP_MATCH, 2);
-
-	{
-		grn_obj *result;
-		result = grn_expr_exec(ctx, expression, 0);
-		if (ctx->rc)
-		{
-			goto exit;
-		}
-		contained = GRN_INT32_VALUE(result) != 0;
-	}
-
-exit:
-	grn_obj_unlink(ctx, expression);
-	GRN_OBJ_FIN(ctx, &buffer);
+	GRN_OBJ_FIN(ctx, &targetBuffer);
+	GRN_OBJ_FIN(ctx, &subTextBuffer);
 
 	return contained;
 }
@@ -1087,24 +1369,28 @@ static void
 PGrnInsert(grn_ctx *ctx,
 		  Relation index,
 		  grn_obj *sourcesTable,
+		  grn_obj *sourcesCtidColumn,
 		  Datum *values,
 		  bool *isnull,
 		  ItemPointer ht_ctid)
 {
 	TupleDesc desc = RelationGetDescr(index);
-	uint64 key = CtidToUInt64(ht_ctid);
 	grn_id id;
 	int i;
 
-	id = grn_table_add(ctx, sourcesTable, &key, sizeof(uint64), NULL);
+	id = grn_table_add(ctx, sourcesTable, NULL, 0, NULL);
+	GRN_UINT64_SET(ctx, &ctidBuffer, CtidToUInt64(ht_ctid));
+	grn_obj_set_value(ctx, sourcesCtidColumn, id, &ctidBuffer, GRN_OBJ_SET);
 
 	for (i = 0; i < desc->natts; i++)
 	{
 		grn_obj *dataColumn;
-		NameData *name = &(desc->attrs[i]->attname);
+		Form_pg_attribute attribute = desc->attrs[i];
+		NameData *name;
 		grn_id domain;
 		unsigned char flags;
 
+		name = &(attribute->attname);
 		if (isnull[i])
 			continue;
 
@@ -1112,7 +1398,7 @@ PGrnInsert(grn_ctx *ctx,
 									name->data, strlen(name->data));
 		domain = PGrnGetType(index, i, &flags);
 		grn_obj_reinit(ctx, &buffer, domain, flags);
-		PGrnGetValue(index, i, &buffer, values[i]);
+		PGrnConvertDatum(values[i], attribute->atttypid, &buffer);
 		grn_obj_set_value(ctx, dataColumn, id, &buffer, GRN_OBJ_SET);
 		grn_obj_unlink(ctx, dataColumn);
 		if (!PGrnCheck("pgroonga: failed to set column value")) {
@@ -1136,19 +1422,121 @@ pgroonga_insert(PG_FUNCTION_ARGS)
 	IndexUniqueCheck checkUnique = PG_GETARG_INT32(5);
 #endif
 	grn_obj *sourcesTable;
+	grn_obj *sourcesCtidColumn;
 
 	sourcesTable = PGrnLookupSourcesTable(index, ERROR);
-	PGrnInsert(ctx, index, sourcesTable, values, isnull, ht_ctid);
+	sourcesCtidColumn = PGrnLookupSourcesCtidColumn(index, ERROR);
+	PGrnInsert(ctx, index, sourcesTable, sourcesCtidColumn,
+			   values, isnull, ht_ctid);
 	grn_db_touch(ctx, grn_ctx_db(ctx));
 
 	PG_RETURN_BOOL(true);
 }
 
+static void
+PGrnFindPrimaryKey(Relation table,
+				   AttrNumber *primaryKeyNumber,
+				   Oid *primaryKeyTypeID)
+{
+	List *indexOIDList;
+	ListCell *cell;
+
+	*primaryKeyNumber = InvalidAttrNumber;
+	*primaryKeyTypeID = InvalidOid;
+
+	indexOIDList = RelationGetIndexList(table);
+	if (indexOIDList == NIL)
+	{
+		return;
+	}
+
+	foreach(cell, indexOIDList)
+	{
+		Oid indexOID = lfirst_oid(cell);
+		Relation index;
+
+		index = index_open(indexOID, NoLock);
+		if (!index->rd_index->indisprimary) {
+			index_close(index, NoLock);
+			continue;
+		}
+
+		if (index->rd_index->indnatts == 1)
+		{
+			TupleDesc desc;
+
+			*primaryKeyNumber = index->rd_index->indkey.values[0];
+
+			desc = RelationGetDescr(table);
+			*primaryKeyTypeID = desc->attrs[*primaryKeyNumber - 1]->atttypid;
+		}
+		index_close(index, NoLock);
+
+		break;
+	}
+
+	list_free(indexOIDList);
+}
+
+static void
+PGrnScanOpaqueInitPrimaryKey(PGrnScanOpaque so, Relation index)
+{
+	Relation table;
+	AttrNumber primaryKeyNumber = InvalidAttrNumber;
+	Oid primaryKeyTypeID = InvalidOid;
+	bool havePrimaryKeyInIndex = false;
+
+	table = RelationIdGetRelation(so->dataTableID);
+	PGrnFindPrimaryKey(table, &primaryKeyNumber, &primaryKeyTypeID);
+
+	if (AttributeNumberIsValid(primaryKeyNumber))
+	{
+		int i, nColumns;
+
+		nColumns = index->rd_index->indkey.ndim;
+		for (i = 0; i < nColumns; i++)
+		{
+			if (index->rd_index->indkey.values[i] == primaryKeyNumber)
+			{
+				havePrimaryKeyInIndex = true;
+				break;
+			}
+		}
+	}
+
+	if (havePrimaryKeyInIndex)
+	{
+		so->primaryKey.number = primaryKeyNumber;
+		so->primaryKey.type = primaryKeyTypeID;
+		so->primaryKey.domain = PGrnGetType(index,
+											primaryKeyNumber,
+											&(so->primaryKey.flags));
+		so->primaryKey.indexColumn = PGrnLookupIndexColumn(index,
+														   primaryKeyNumber - 1,
+														   ERROR);
+		so->primaryKey.lexicon =
+			grn_ctx_at(ctx, so->primaryKey.indexColumn->header.domain);
+	}
+	else
+	{
+		so->primaryKey.number = InvalidAttrNumber;
+		so->primaryKey.type = InvalidOid;
+		so->primaryKey.domain = GRN_ID_NIL;
+		so->primaryKey.flags = 0;
+		so->primaryKey.lexicon = NULL;
+		so->primaryKey.indexColumn = NULL;
+	}
+
+	RelationClose(table);
+}
+
 static void
 PGrnScanOpaqueInit(PGrnScanOpaque so, Relation index)
 {
 	so->dataTableID = index->rd_index->indrelid;
+	PGrnScanOpaqueInitPrimaryKey(so, index);
 	so->sourcesTable = PGrnLookupSourcesTable(index, ERROR);
+	so->sourcesCtidColumn = PGrnLookupSourcesCtidColumn(index, ERROR);
 	GRN_VOID_INIT(&(so->minBorderValue));
 	GRN_VOID_INIT(&(so->maxBorderValue));
 	so->searched = NULL;
@@ -1156,7 +1544,7 @@ PGrnScanOpaqueInit(PGrnScanOpaque so, Relation index)
 	so->targetTable = NULL;
 	so->indexCursor = NULL;
 	so->tableCursor = NULL;
-	so->keyAccessor = NULL;
+	so->ctidAccessor = NULL;
 	so->scoreAccessor = NULL;
 	so->currentID = GRN_ID_NIL;
 
@@ -1172,10 +1560,10 @@ PGrnScanOpaqueReinit(PGrnScanOpaque so)
 		grn_obj_unlink(ctx, so->scoreAccessor);
 		so->scoreAccessor = NULL;
 	}
-	if (so->keyAccessor)
+	if (so->ctidAccessor)
 	{
-		grn_obj_unlink(ctx, so->keyAccessor);
-		so->keyAccessor = NULL;
+		grn_obj_unlink(ctx, so->ctidAccessor);
+		so->ctidAccessor = NULL;
 	}
 	if (so->indexCursor)
 	{
@@ -1288,6 +1676,7 @@ PGrnSearchBuildConditions(IndexScanDesc scan,
 	for (i = 0; i < scan->numberOfKeys; i++)
 	{
 		ScanKey key = &(scan->keyData[i]);
+		Form_pg_attribute attribute;
 		grn_bool isValidStrategy = GRN_TRUE;
 		const char *targetColumnName;
 		grn_obj *targetColumn;
@@ -1298,11 +1687,13 @@ PGrnSearchBuildConditions(IndexScanDesc scan,
 		if (key->sk_flags & SK_ISNULL)
 			continue;
 
+		attribute = desc->attrs[key->sk_attno - 1];
+
 		GRN_EXPR_CREATE_FOR_QUERY(ctx, so->sourcesTable,
 								  matchTarget, matchTargetVariable);
 		GRN_PTR_PUT(ctx, &(data->matchTargets), matchTarget);
 
-		targetColumnName = desc->attrs[key->sk_attno - 1]->attname.data;
+		targetColumnName = attribute->attname.data;
 		targetColumn = grn_obj_column(ctx, so->sourcesTable,
 									  targetColumnName,
 									  strlen(targetColumnName));
@@ -1315,7 +1706,19 @@ PGrnSearchBuildConditions(IndexScanDesc scan,
 			domain = PGrnGetType(index, key->sk_attno - 1, NULL);
 			grn_obj_reinit(ctx, &buffer, domain, flags);
 		}
-		PGrnGetQuery(index, key->sk_attno - 1, &buffer, key->sk_argument);
+		{
+			Oid valueTypeID = attribute->atttypid;
+			switch (valueTypeID)
+			{
+			case VARCHARARRAYOID:
+				valueTypeID = VARCHAROID;
+				break;
+			case TEXTARRAYOID:
+				valueTypeID = TEXTOID;
+				break;
+			}
+			PGrnConvertDatum(key->sk_argument, valueTypeID, &buffer);
+		}
 
 		switch (key->sk_strategy)
 		{
@@ -1493,9 +1896,9 @@ PGrnOpenTableCursor(IndexScanDesc scan, ScanDirection dir)
 	so->tableCursor = grn_table_cursor_open(ctx, table,
 											NULL, 0, NULL, 0,
 											offset, limit, flags);
-	so->keyAccessor = grn_obj_column(ctx, table,
-									 GRN_COLUMN_NAME_KEY,
-									 GRN_COLUMN_NAME_KEY_LEN);
+	so->ctidAccessor = grn_obj_column(ctx, table,
+									  PGrnSourcesCtidColumnName,
+									  PGrnSourcesCtidColumnNameLength);
 	if (so->searched)
 	{
 		so->scoreAccessor = grn_obj_column(ctx, so->searched,
@@ -1545,20 +1948,26 @@ PGrnFillBorder(IndexScanDesc scan,
 			   int *flags)
 {
 	Relation index = scan->indexRelation;
+	TupleDesc desc;
 	PGrnScanOpaque so = (PGrnScanOpaque) scan->opaque;
 	grn_obj *minBorderValue;
 	grn_obj *maxBorderValue;
 	int i;
 
+	desc = RelationGetDescr(index);
+
 	minBorderValue = &(so->minBorderValue);
 	maxBorderValue = &(so->maxBorderValue);
 	for (i = 0; i < scan->numberOfKeys; i++)
 	{
 		ScanKey key = &(scan->keyData[i]);
 		AttrNumber attrNumber;
+		Form_pg_attribute attribute;
 		grn_id domain;
 
 		attrNumber = key->sk_attno - 1;
+		attribute = desc->attrs[attrNumber];
+
 		domain = PGrnGetType(index, attrNumber, NULL);
 		switch (key->sk_strategy)
 		{
@@ -1567,7 +1976,9 @@ PGrnFillBorder(IndexScanDesc scan,
 			if (maxBorderValue->header.type != GRN_DB_VOID)
 			{
 				grn_obj_reinit(ctx, &buffer, domain, 0);
-				PGrnGetQuery(index, attrNumber, &buffer, key->sk_argument);
+				PGrnConvertDatum(key->sk_argument,
+								 attribute->atttypid,
+								 &buffer);
 				if (!PGrnIsMeaningfullMaxBorderValue(maxBorderValue,
 													 &buffer,
 													 *flags,
@@ -1577,7 +1988,9 @@ PGrnFillBorder(IndexScanDesc scan,
 				}
 			}
 			grn_obj_reinit(ctx, maxBorderValue, domain, 0);
-			PGrnGetQuery(index, attrNumber, maxBorderValue, key->sk_argument);
+			PGrnConvertDatum(key->sk_argument,
+							 attribute->atttypid,
+							 maxBorderValue);
 			*max = GRN_BULK_HEAD(maxBorderValue);
 			*maxSize = GRN_BULK_VSIZE(maxBorderValue);
 			*flags &= ~(GRN_CURSOR_LT | GRN_CURSOR_LE);
@@ -1595,8 +2008,9 @@ PGrnFillBorder(IndexScanDesc scan,
 			if (minBorderValue->header.type != GRN_DB_VOID)
 			{
 				grn_obj_reinit(ctx, &buffer, domain, 0);
-				PGrnGetQuery(index, attrNumber, &buffer,
-							 key->sk_argument);
+				PGrnConvertDatum(key->sk_argument,
+								 attribute->atttypid,
+								 &buffer);
 				if (!PGrnIsMeaningfullMinBorderValue(minBorderValue,
 													 &buffer,
 													 *flags,
@@ -1606,7 +2020,9 @@ PGrnFillBorder(IndexScanDesc scan,
 				}
 			}
 			grn_obj_reinit(ctx, minBorderValue, domain, 0);
-			PGrnGetQuery(index, attrNumber, minBorderValue, key->sk_argument);
+			PGrnConvertDatum(key->sk_argument,
+							 attribute->atttypid,
+							 minBorderValue);
 			*min = GRN_BULK_HEAD(minBorderValue);
 			*minSize = GRN_BULK_VSIZE(minBorderValue);
 			*flags &= ~(GRN_CURSOR_GT | GRN_CURSOR_GE);
@@ -1674,9 +2090,9 @@ PGrnRangeSearch(IndexScanDesc scan, ScanDirection dir)
 											indexCursorMin,
 											indexCursorMax,
 											indexCursorFlags);
-	so->keyAccessor = grn_obj_column(ctx, so->sourcesTable,
-									 GRN_COLUMN_NAME_KEY,
-									 GRN_COLUMN_NAME_KEY_LEN);
+	so->ctidAccessor = grn_obj_column(ctx, so->sourcesTable,
+									  PGrnSourcesCtidColumnName,
+									  PGrnSourcesCtidColumnNameLength);
 }
 
 static bool
@@ -1753,12 +2169,7 @@ pgroonga_gettuple(PG_FUNCTION_ARGS)
 
 	if (scan->kill_prior_tuple && so->currentID != GRN_ID_NIL)
 	{
-		grn_obj key;
-		GRN_UINT64_INIT(&key, 0);
-		grn_obj_get_value(ctx, so->keyAccessor, so->currentID, &key);
-		grn_table_delete(ctx, so->sourcesTable,
-						 GRN_BULK_HEAD(&key), GRN_BULK_VSIZE(&key));
-		GRN_OBJ_FIN(ctx, &key);
+		grn_table_delete_by_id(ctx, so->sourcesTable, so->currentID);
 	}
 
 	if (so->indexCursor)
@@ -1782,12 +2193,9 @@ pgroonga_gettuple(PG_FUNCTION_ARGS)
 	}
 	else
 	{
-		grn_obj key;
-
-		GRN_UINT64_INIT(&key, 0);
-		grn_obj_get_value(ctx, so->keyAccessor, so->currentID, &key);
-		scan->xs_ctup.t_self = UInt64ToCtid(GRN_UINT64_VALUE(&key));
-		GRN_OBJ_FIN(ctx, &key);
+		GRN_BULK_REWIND(&ctidBuffer);
+		grn_obj_get_value(ctx, so->ctidAccessor, so->currentID, &ctidBuffer);
+		scan->xs_ctup.t_self = UInt64ToCtid(GRN_UINT64_VALUE(&ctidBuffer));
 
 		PG_RETURN_BOOL(true);
 	}
@@ -1804,11 +2212,9 @@ pgroonga_getbitmap(PG_FUNCTION_ARGS)
 	TIDBitmap *tbm = (TIDBitmap *) PG_GETARG_POINTER(1);
 	PGrnScanOpaque so = (PGrnScanOpaque) scan->opaque;
 	int64 nRecords = 0;
-	grn_obj key;
 
 	PGrnEnsureCursorOpened(scan, ForwardScanDirection);
 
-	GRN_UINT64_INIT(&key, 0);
 	if (so->indexCursor)
 	{
 		grn_posting *posting;
@@ -1816,9 +2222,9 @@ pgroonga_getbitmap(PG_FUNCTION_ARGS)
 		while ((posting = grn_index_cursor_next(ctx, so->indexCursor, &termID)))
 		{
 			ItemPointerData ctid;
-			GRN_BULK_REWIND(&key);
-			grn_obj_get_value(ctx, so->keyAccessor, posting->rid, &key);
-			ctid = UInt64ToCtid(GRN_UINT64_VALUE(&key));
+			GRN_BULK_REWIND(&ctidBuffer);
+			grn_obj_get_value(ctx, so->ctidAccessor, posting->rid, &ctidBuffer);
+			ctid = UInt64ToCtid(GRN_UINT64_VALUE(&ctidBuffer));
 			tbm_add_tuples(tbm, &ctid, 1, false);
 			nRecords++;
 		}
@@ -1829,14 +2235,13 @@ pgroonga_getbitmap(PG_FUNCTION_ARGS)
 		while ((id = grn_table_cursor_next(ctx, so->tableCursor)) != GRN_ID_NIL)
 		{
 			ItemPointerData ctid;
-			GRN_BULK_REWIND(&key);
-			grn_obj_get_value(ctx, so->keyAccessor, id, &key);
-			ctid = UInt64ToCtid(GRN_UINT64_VALUE(&key));
+			GRN_BULK_REWIND(&ctidBuffer);
+			grn_obj_get_value(ctx, so->ctidAccessor, id, &ctidBuffer);
+			ctid = UInt64ToCtid(GRN_UINT64_VALUE(&ctidBuffer));
 			tbm_add_tuples(tbm, &ctid, 1, false);
 			nRecords++;
 		}
 	}
-	GRN_OBJ_FIN(ctx, &key);
 
 	PG_RETURN_INT64(nRecords);
 }
@@ -1890,8 +2295,8 @@ PGrnBuildCallback(Relation index,
 	PGrnBuildState bs = (PGrnBuildState) state;
 
 	if (tupleIsAlive) {
-		PGrnInsert(ctx, index, bs->sourcesTable, values,
-				   isnull, &(htup->t_self));
+		PGrnInsert(ctx, index, bs->sourcesTable, bs->sourcesCtidColumn,
+				   values, isnull, &(htup->t_self));
 		bs->nIndexedTuples++;
 	}
 }
@@ -1921,7 +2326,10 @@ pgroonga_build(PG_FUNCTION_ARGS)
 	GRN_PTR_INIT(&lexicons, GRN_OBJ_VECTOR, GRN_ID_NIL);
 	PG_TRY();
 	{
-		PGrnCreate(index, &(bs.sourcesTable), &lexicons);
+		PGrnCreate(index,
+				   &(bs.sourcesTable),
+				   &(bs.sourcesCtidColumn),
+				   &lexicons);
 		nHeapTuples = IndexBuildHeapScan(heap, index, indexInfo, true,
 										 PGrnBuildCallback, &bs);
 		PGrnSetSources(index, bs.sourcesTable);
@@ -1961,12 +2369,13 @@ pgroonga_buildempty(PG_FUNCTION_ARGS)
 {
 	Relation index = (Relation) PG_GETARG_POINTER(0);
 	grn_obj *sourcesTable = NULL;
+	grn_obj *sourcesCtidColumn = NULL;
 	grn_obj lexicons;
 
 	GRN_PTR_INIT(&lexicons, GRN_OBJ_VECTOR, GRN_ID_NIL);
 	PG_TRY();
 	{
-		PGrnCreate(index, &sourcesTable, &lexicons);
+		PGrnCreate(index, &sourcesTable, &sourcesCtidColumn, &lexicons);
 		PGrnSetSources(index, sourcesTable);
 	}
 	PG_CATCH();
@@ -2044,15 +2453,19 @@ pgroonga_bulkdelete(PG_FUNCTION_ARGS)
 
 	PG_TRY();
 	{
-		while (grn_table_cursor_next(ctx, cursor) != GRN_ID_NIL)
+		grn_id id;
+		grn_obj *sourcesCtidColumn;
+
+		sourcesCtidColumn = PGrnLookupSourcesCtidColumn(index, ERROR);
+		while ((id = grn_table_cursor_next(ctx, cursor)) != GRN_ID_NIL)
 		{
 			ItemPointerData	ctid;
-			uint64 *key;
 
 			CHECK_FOR_INTERRUPTS();
 
-			grn_table_cursor_get_key(ctx, cursor, (void **) &key);
-			ctid = UInt64ToCtid(*key);
+			GRN_BULK_REWIND(&ctidBuffer);
+			grn_obj_get_value(ctx, sourcesCtidColumn, id, &ctidBuffer);
+			ctid = UInt64ToCtid(GRN_UINT64_VALUE(&ctidBuffer));
 			if (callback(&ctid, callback_state))
 			{
 				grn_table_cursor_delete(ctx, cursor);
@@ -2159,11 +2572,41 @@ pgroonga_vacuumcleanup(PG_FUNCTION_ARGS)
 Datum
 pgroonga_costestimate(PG_FUNCTION_ARGS)
 {
-	/*
-	 * We cannot use genericcostestimate because it is a static funciton.
-	 * Use gistcostestimate instead, which just calls genericcostestimate.
-	 */
 	return gistcostestimate(fcinfo);
+
+#ifdef NOT_USED
+	PlannerInfo *root = (PlannerInfo *) PG_GETARG_POINTER(0);
+	IndexPath *path = (IndexPath *) PG_GETARG_POINTER(1);
+	double loopCount = PG_GETARG_FLOAT8(2);
+	Cost *indexStartupCost = (Cost *) PG_GETARG_POINTER(3);
+	Cost *indexTotalCost = (Cost *) PG_GETARG_POINTER(4);
+	Selectivity *indexSelectivity = (Selectivity *) PG_GETARG_POINTER(5);
+	double *indexCorrelation = (double *) PG_GETARG_POINTER(6);
+	IndexOptInfo *index = path->indexinfo;
+
+	/* TODO: Use more clever logic.
+	 *
+	 * We want to use index scan rather than bitmap scan for full text search.
+	 * Because bitmap scan requires bitmap heap scan that is slow for
+	 * large result set.
+	 *
+	 * We want to use bitmap scan rather than index scan for OR search.
+	 * Because we can't use index scan for OR search.
+	 *
+	 * We want to use the default scan for other cases.
+	 */
+	*indexSelectivity = clauselist_selectivity(root,
+											   path->indexquals,
+											   index->rel->relid,
+											   JOIN_INNER,
+											   NULL);
+
+	*indexStartupCost = 0.0;
+	*indexTotalCost = 0.0;
+	*indexCorrelation = 0.0;
+
+	PG_RETURN_VOID();
+#endif
 }
 
 /**
